import os
import copy
import json
import time
import asyncio
import aiofiles
import contextlib
import click

from ..common.utils import run_blocking_func_in_executor

# The default config is the basic configuration of the netunnel.
# In case the configuration already exists, it's performing an update over the default config.
# This way, in case the default configuration has a new key in some version, it will be added automatically.
# We don't performing a full recursive merge between the current and the default configuration to protect
# dict values like 'allowed_tunnel_destination' from being changed unexpectedly.
_DEFAULT_CONFIG = {
    'allowed_tunnel_destinations': {"127.0.0.1": "*"},
    'http_proxy': None,
    'http_proxy_test_url': 'https://google.com',
    'peers': [],
    'allow_unverified_ssl_peers': False,
    'revision': 1
    # secret_key is not provided here so it won't be dynamically stored in the configuration file by mistake
}
ENV_VARIABLES_PREFIX = 'NETUNNEL_'


def get_default_config(use_env_vars=True) -> dict:
    """
    If use_env_vars is True, config values will be override by environment variables if defined.
    For example, to override the value of "allowed_tunnel_destinations", export "NETUNNEL_ALLOWED_TUNNEL_DESTINATIONS".
    Environment variables are expected to be in json format, and are usually defined in the systemd service.
    """
    default_config = copy.deepcopy(_DEFAULT_CONFIG)
    if use_env_vars:
        for key in default_config:
            env_var_name = ENV_VARIABLES_PREFIX + key.upper()
            if env_var_name in os.environ:
                default_config[key] = json.loads(os.environ[env_var_name])
    return default_config


class NETunnelConfiguration:
    def __init__(self, config_path=None):
        """
        USE ONLY `await NETunnelConfiguration.create(path)` TO INITIALIZE AN OBJECT

        Manage configurations for NETunnelServer in JSON format.
        Store configurations in memory unless config_path was given.
        :param config_path: Optional path to a configuration file in which to store changes.
        """
        self._config_path = config_path
        # This lock is used to prevent multiple disk writes when saving the configuration using safe writes.
        self._saving_config_lock = asyncio.Lock()
        self._config = get_default_config()

    @classmethod
    async def create(cls, config_path=None):
        self = NETunnelConfiguration(config_path=config_path)
        await self._initialize()
        return self

    async def recreate(self):
        """
        Recreate the configuration file from the default settings.
        This will erase the current configurations.
        """
        self._config = get_default_config()
        await self.save()

    async def _initialize(self):
        """
        Initialize config by loading config_path
        """
        if self._config_path:
            with contextlib.suppress(FileNotFoundError):
                async with aiofiles.open(self._config_path) as config_file:
                    data = await config_file.read()
                    self._config.update(json.loads(data))
        await self.save()

    async def save(self):
        """
        Save configurations using safe writes to the config path if exists
        """
        if self._config_path is None:
            return
        async with self._saving_config_lock:
            temp_config_path = f"{self._config_path}.{time.time()}"
            # First we write all the config to a temporary file
            async with aiofiles.open(temp_config_path, 'w') as temp_config_file:
                await temp_config_file.write(json.dumps(self._config, indent=4))
                await temp_config_file.flush()
                await run_blocking_func_in_executor(os.fsync, temp_config_file)
            # Now we overwrite the config file with an atomic operation
            await run_blocking_func_in_executor(os.rename, temp_config_path, self._config_path)

    def __getitem__(self, key):
        return self._config[key]

    def __setitem__(self, key, value):
        self._config[key] = value

    def __contains__(self, key):
        return key in self._config

    def get(self, key, default=None):
        return self._config.get(key, default)


@click.group()
def main():
    pass


@main.command()
@click.argument('path')
@click.option('--custom-changes', default='{}', help='A json dump string of the changes you want to make in the default configuration')
def create(path, custom_changes):
    custom_changes = json.loads(custom_changes)

    config_dir = os.path.dirname(path)
    os.makedirs(config_dir, exist_ok=True)
    with open(path, 'w') as f:
        json.dump(custom_changes, f)

    async def create_config():
        config = await NETunnelConfiguration.create(path)
        await config.save()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(create_config())


if __name__ == '__main__':
    main()



from typing import Union, Callable, Coroutine, Any, Dict, List
from .utils import EventClientWebSocketResponse, EventWebSocketResponse
from aiohttp import ClientWebSocketResponse
from aiohttp.web import WebSocketResponse


# Types
WebsocketType = Union[EventClientWebSocketResponse, EventWebSocketResponse, WebSocketResponse, ClientWebSocketResponse]
CallableOfCoroutine = Callable[..., Coroutine]
TunnelId = int
ChannelId = int


# Constants
CLIENT_CHANNEL_HEARTBEAT = 20
SERVER_MAXIMUM_CLIENT_CHANNELS = 1000
SERVER_MAXIMUM_CLIENT_CHANNELS_ERROR = 'Reached maximum allowed number of control_channels'
MIN_STATIC_TUNNEL_LOCAL_PORT = 20000
MAX_STATIC_TUNNEL_LOCAL_PORT = 21000


from typing import TYPE_CHECKING
from aiohttp import web
if TYPE_CHECKING:
    from ..client import NETunnelClient

import abc


class NETunnelServerAuth(abc.ABC):
    def __init__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    async def get_client_for_peer(self, *args, **kwargs) -> 'NETunnelClientAuth':
        """
        Returns an instance of NETunnelClientAuth which will be used by a peer object to authenticate with it.
        The parameters of the subclass should be the expected parameters of the constructor of your NETunnelClientAuth class:
        for example:

        async def get_client_for_peer(self, username, password):
            return NETunnelClientAuth(username, password)
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def is_authenticated(self, request: web.Request) -> bool:
        """
        Return True if the request is authorized and False if not.
        The headers of the request should include the return value of NETunnelClientAuth.get_authorized_headers()
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def authenticate(self, request: web.Request):
        """
        Perform the server-side authentication. This method handling the route /authenticate on the NETunnel server.

        The NETunnelClientAuth.authenticate(client) should make a request to the /authenticate uri of the NETunnel server,
        and this method will be called as the route handler.
        """
        raise NotImplementedError


class NETunnelClientAuth(abc.ABC):
    def __init__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    async def authenticate(self, client: 'NETunnelClient', *args, **kwargs):
        """
        Perform the client-side authentication. The NETunnel server is exposing the /authenticate uri which will
        be handled by NETunnelServerAuth.authenticate(request).
        Raises NETunnelAuthError if failed to authenticate.

        After calling this method, self.get_authorized_headers is expected to return valid headers which will
        be authorized by the remote.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def is_authenticated(self) -> bool:
        """
        Return True if we are authenticated with the remote.
        This should return False if the tokens returned by the server are expired.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def get_authorized_headers(self) -> dict:
        """
        Return headers which can be used to make authorized requests to the remote server.
        Raises NETunnelNotAuthenticatedError if we need to authenticate first/again
        """
        raise NotImplementedError

    @abc.abstractmethod
    def dump_object(self) -> dict:
        """
        This method is not async since we're using marshmallow to parse this data and it does not support asyncio yet.
        Return a dump of this NETunnelClientAuth instance's data so it can be stored in the configuration
        and reloaded during startup. This is used by NETunnelServer when handling peer objects.
        The return value should be a dictionary that can be given to NETunnelServerAuth.get_client_for_peer() and
        will generate a valid NETunnelClientAuth object.
        """
        raise NotImplementedError


class ServerNoAuth(NETunnelServerAuth):
    async def get_client_for_peer(self, *args, **kwargs) -> 'NETunnelClientAuth':
        return ClientNoAuth(*args, **kwargs)

    async def is_authenticated(self, request: web.Request):
        return True

    async def authenticate(self, request: web.Request):
        pass


class ClientNoAuth(NETunnelClientAuth):
    async def authenticate(self, client: 'NETunnelClient', *args, **kwargs):
        pass

    async def is_authenticated(self):
        return True

    async def get_authorized_headers(self):
        return {}

    def dump_object(self):
        return {}



"""
Utilities for creating tunnels on top of websockets
The 2 classes you'll need to use are:
InputTunnel -> Listen on a local address and port. For each connection, it takes a new websocket and forward the data on top of it
OutputTunnel -> Listen on a websocket for incoming messages and forward all data to a designated local address and port

Diagram for how traffic looks like inside the tunnel:

Connection1 ->                  Websocket1 ->               ->
Connection2 ->  InputTunnel ->  Websocket2 ->  OutputTunnel -> Service
Connection3 ->                  Websocket3 ->               ->

traffic goes bi-directional to allow `Service` to respond

For 'reverse tunnel', you just need to create them in reverse.
"""
import asyncio
import logging
import subprocess
import contextlib
import abc
import aiohttp
import aiohttp.web

from typing import List, Coroutine, Callable
from asyncio.base_events import Server
from .const import WebsocketType, CallableOfCoroutine
from .utils import get_logger, EventQueue, object_in_list, task_in_list_until_done, asyncio_current_task

CONNECTION_READ_CHUNK_SIZE = 65535
# This message is used when a connection arrived on InputTunnel to send to the OutputTunnel.
# This is utilized by OutputTunnel to avoid establishing connections with websockets that aren't
# needed yet by InputTunnel and just exists to reduce latency for new connections.
WEBSOCKET_CONNECTION_START_MESSAGE = 'pasten'


class _ConnectionHandler:
    def __init__(self, websocket: WebsocketType, connection_reader: asyncio.StreamReader, connection_writer: asyncio.StreamWriter, logger: logging.Logger):
        """
        Manage a connection. This is used internally, do not initialize this.
        Used by both InputTunnel for new connections to the tunnel, and OutputTunnel for connections established to a service.
        Create read and write pipes between the websocket and the connection:  websocket <-> connection
        :param websocket: websocket to receive and send data in and from the connection
        :param connection_reader: StreamReader object used to read data and send it to the websocket
        :param connection_writer: StreamWriter object used to write data that came from the websocket
        """
        self._websocket = websocket
        self._connection_reader = connection_reader
        self._connection_writer = connection_writer
        self._websocket_to_connection_task: asyncio.Task = None
        self._connection_to_websocket_task: asyncio.Task = None
        self._logger = logger
        # The following event is triggered when either the websocket or the connection closed or reached end-of-file
        self._eof_event = asyncio.Event()
        # Replace with self._connection_writer.wait_closed in Python3.7
        self._shutdown_event = asyncio.Event()

    async def health_check(self):
        """
        Check that both the websocket and the connection can still stream data
        """
        if self._websocket_to_connection_task and self._connection_to_websocket_task:
            return not self._websocket_to_connection_task.done() and not self._connection_to_websocket_task.done() and \
                   not self._websocket.closed and not self._connection_reader.at_eof()
        return False

    async def _websocket_to_connection(self):
        """
        This method stream data that comes from the websocket to the connection_writer.
        If either there is not data left to send (the websocket closed) or there was an error in the communication,
        we signal the self._eof_event.
        """
        try:
            async for msg in self._websocket:
                if msg.type == aiohttp.WSMsgType.BINARY:
                    self._connection_writer.write(msg.data)
                    try:
                        await self._connection_writer.drain()
                    except ConnectionError:
                        self._logger.debug('Connection closed. cannot send data from websocket. Closing websocket-to-connection pipe')
                        break
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    self._logger.exception("Websocket disconnected unexpectedly", exc_info=self._websocket.exception())
                    break
                else:
                    self._logger.warning('unexpected message received from server. Type: %s, Data: %s', msg.type.name, msg.data)
        finally:
            self._eof_event.set()

    async def _connection_to_websocket(self):
        """
        This method stream data that comes from the connection_reader to the websocket.
        If there is no more data to read or there was exception in either reading or writing to the websocket,
        we signal the self._eof_event.
        """
        try:
            while not self._connection_reader.at_eof() and not self._websocket.closed:
                data = await self._connection_reader.read(CONNECTION_READ_CHUNK_SIZE)
                try:
                    await self._websocket.send_bytes(data)
                except ConnectionError:
                    self._logger.debug('Websocket closed. cannot send data from connection. Closing connection-to-websocket pipe')
                    break
        finally:
            self._eof_event.set()

    async def _start(self):
        """
        Start read and write pipes
        This should not be used directly, use self.run_until_eof() instead.
        """
        self._shutdown_event.clear()
        self._websocket_to_connection_task = asyncio.ensure_future(self._websocket_to_connection())
        self._connection_to_websocket_task = asyncio.ensure_future(self._connection_to_websocket())

    async def run_until_eof(self):
        """
        Start working and wait until either websocket's closed (remote at eof) or until connection at eof
        """
        await self._start()
        await self._eof_event.wait()
        await self.stop()

    async def stop(self):
        """
        Stop read and write pipes and close the connection
        """
        for task in (self._websocket_to_connection_task, self._connection_to_websocket_task):
            if task:
                task.cancel()
                try:
                    await task
                except asyncio.CancelledError:
                    pass
                except Exception:
                    self._logger.exception('An error occurred while awaiting connection pipe to be closed:')
        self._connection_writer.close()
        self._shutdown_event.set()
    
    async def join(self):
        with contextlib.suppress(asyncio.CancelledError):
            await self._shutdown_event.wait()


class Tunnel(abc.ABC):
    def __init__(self, entrance_address=None, entrance_port=None, exit_address=None, exit_port=None,
                 logger: logging.Logger = None, stop_tunnel_on_remote_callback: Callable[[bool], Coroutine] = None,
                 stopped_callback: CallableOfCoroutine = None):
        """
        This is an interface to manage tunnels.
        A tunnel is an InputTunnel on one side and OutputTunnel on the other side, and reverse tunnel is just the same but
        in reverse. We want the user to have the same interface for when using both normal tunnels and reverse tunnels,
        therefore this class will be used to simplify his usage, and also to share common utilities between InputTunnel and OutputTunnel.
        :param entrance_address: Tunnel's entrance address
        :param entrance_port: Tunnel's entrance port
        :param exit_address: Tunnel's exit address
        :param exit_port: Tunnel's exit port
        :param logger: A logging.Logger object
        :param stop_tunnel_on_remote_callback: callback for when we need to stop the remote tunnel. Accept parameter `force`
        :param stopped_callback: callback for after the tunnel is stopped
        """
        self._entrance_address = entrance_address
        self._entrance_port = entrance_port
        self._exit_address = exit_address
        self._exit_port = exit_port
        self._connections: List[_ConnectionHandler] = []
        self._logger = logger or get_logger(__name__)
        self._stop_tunnel_on_remote_callback = stop_tunnel_on_remote_callback
        self._stopped_callback = stopped_callback
        # This queue is responsible to hold new websockets ready to be used for incoming connections.
        # Each item on the queue is a tuple of (websocket, websocket_done_event):
        # websocket - this websocket is connected to the tunnel's other side
        # websocket_done_event - this event is triggered by the tunnel when the websocket is no needed anymore because eof reached
        # This helps us to separate the tunnel logic from the client/server logic and also reduce latency in creating websockets.
        self._websocket_queue = EventQueue(maxsize=1)
        self._running = False
        self._shutdown = asyncio.Event()

    def get_entrance_socket(self):
        return self._entrance_address, self._entrance_port

    def get_exit_socket(self):
        return self._exit_address, self._exit_port

    @property
    def running(self):
        return self._running

    async def health_check(self):
        """
        Perform a health check to the tunnel.
        Make sure that the tunnel is running and all the existing connections are working.
        see InputTunnel and OutputTunnel for their individual checks as well.
        """
        return self.running and all([await connection.health_check() for connection in self._connections])

    async def start(self):
        self._running = True
        self._shutdown.clear()

    async def stop(self, stop_remote_tunnel=True, force=False):
        """
        Stop the tunnel by closing the connections and the queued websocket.
        We first initiate the queued websocket to be closed in the background so it will start listening for messages,
        and then we ask the remote to also stop so it will start listening to his queued websocket and receive our close message.
        :param stop_remote_tunnel: Whether to call the callback for stopping tunnel on remote. Remote won't be using it to prevent a loop.
        :param force: Whether to close connections forcefully. Otherwise, await for the connections to be closed normally
        """
        if self._running is False:
            return
        self._running = False
        websocket: WebsocketType = None
        try:
            while self._connections:
                    connection = self._connections.pop()
                    if force:
                        await connection.stop()
                    else:
                        await connection.join()
            with contextlib.suppress(asyncio.QueueEmpty):
                websocket = await self._get_websocket(nowait=True)
                websocket.close_nowait()
            if stop_remote_tunnel and self._stop_tunnel_on_remote_callback:
                await self._stop_tunnel_on_remote_callback(force=force)
        except Exception:
            self._logger.exception('An error occurred while stopping tunnel:')
            raise
        finally:
            # stopping tunnel on remote might throw a network related exception but tunnel is still considered close and
            # we still need to await for the yet-to-be-closed websocket to be closed cleanly (probably by timeout if there were an exception)
            try:
                if websocket is not None:
                    await websocket.wait_closed()
            except aiohttp.WebSocketError:
                self._logger.exception('An error occurred while closing an unused websocket:')
            self._shutdown.set()
            if self._stopped_callback:
                await self._stopped_callback()

    async def join(self):
        """
        Blocks until the tunnel is closed
        """
        await self._shutdown.wait()

    async def feed_websocket(self, websocket: WebsocketType):
        """
        Put a websocket in the self._websocket_queue
        """
        if not self.running:
            raise RuntimeError('Tunnel is not running')
        await self._websocket_queue.put(websocket)

    async def wait_new_websocket(self):
        """
        Blocks until there is a new websocket on the queue
        """
        await self._websocket_queue.join_no_empty()

    async def wait_no_websocket(self):
        """
        Blocks until there is no websocket on the queue
        """
        await self._websocket_queue.join()

    async def _get_websocket(self, nowait=False) -> WebsocketType:
        """
        Return an item from self._websocket_queue, and also mark the task as done.
        Marking done is for the client/server to already start working on the next websocket so we can
        reduce latency for new connections.
        :param nowait: Call get_nowait instead of get. This won't block but raise if no websocket on the queue
        """
        if nowait:
            websocket = self._websocket_queue.get_nowait()
        else:
            websocket = await self._websocket_queue.get()
        self._websocket_queue.task_done()
        return websocket


class InputTunnel(Tunnel):
    def __init__(self, entrance_address, entrance_port, websocket_feeder_coro: Coroutine, exit_address=None, exit_port=None,
                 logger: logging.Logger = None, stop_tunnel_on_remote_callback: Callable[[bool], Coroutine] = None, stopped_callback: CallableOfCoroutine = None):
        """
        Creates a local tcp server and listen on _entrance_address and _entrance_port
        For each connection, it creates a websocket to the remote server and start streaming the data on top of it and vice-versa
        This is basically the tunnel's entrance while OutputTunnel is the exit
        :param websocket_feeder_coro: coroutine which responsible to feed this InputTunnel instance with websockets.
                                      it makes the code simpler since InputTunnel can hook it to the start and close methods
        """
        super().__init__(entrance_address, entrance_port, exit_address, exit_port, logger, stop_tunnel_on_remote_callback, stopped_callback)
        self._websocket_feeder_coro: Coroutine = websocket_feeder_coro
        self._websocket_feeder_task: asyncio.Future = None
        # A list of the active tasks which serves connections to self._server. Used for cleanup when stopping the tunnel
        self._active_handle_connection_tasks = []
        self._server: Server = None

    async def health_check(self):
        # According to asyncio doc, Server.sockets is None if the server is closed
        # On python3.7+ replace self._server.sockets with self._server.is_serving()
        return await super().health_check() and self._server.sockets and not self._websocket_feeder_task.done()

    def get_local_socket(self):
        return self.get_entrance_socket()

    def get_remote_socket(self):
        return self.get_exit_socket()

    async def start(self):
        """
        Start the server to listen and handle incoming connections from the local address and port.
        This also starts the websocket_feeder so we can start to receive new websockets
        """
        self._server = await asyncio.start_server(self._handle_connection, host=self._entrance_address, port=self._entrance_port)
        self._entrance_address, self._entrance_port = self._server.sockets[0].getsockname()
        self._logger.debug('Start listening on %s:%s', self._entrance_address, self._entrance_port)
        await super().start()
        self._websocket_feeder_task = asyncio.ensure_future(self._websocket_feeder_coro)

    async def stop(self, stop_remote_tunnel=True, force=False):
        if self._running is False:
            return
        try:
            self._logger.debug('Stop listening on %s:%s', self._entrance_address, self._entrance_port)
            self._server.close()
            await self._server.wait_closed()
            if force:
                self._logger.debug('Stopping any remaining connections on %s:%s', self._entrance_address, self._entrance_port)
                for task in self._active_handle_connection_tasks:
                    task.cancel()
            self._logger.debug('Stopping websocket feeder on %s:%s', self._entrance_address, self._entrance_port)
            self._websocket_feeder_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._websocket_feeder_task
        except Exception:
            self._logger.exception('An error occurred while stopping tunnel entrance handler:')
        finally:
            await super().stop(stop_remote_tunnel=stop_remote_tunnel, force=force)
            # Any exception will be handled or at least logged in the task, yet we await it to satisfy asyncio.
            # We await only after super().stop(...) since we need self.running to be False.
            with contextlib.suppress(Exception):
                await asyncio.gather(*self._active_handle_connection_tasks)

    @staticmethod
    def _get_connection_identity_display_name(connection_writer: asyncio.StreamWriter) -> str:
        """
        Try to query the connection_writer to identify the connection source and return it.
        If we failed to identify the source, we return "Unknown"
        """
        client_process: subprocess.Popen = connection_writer.get_extra_info('subprocess')
        client_peername = connection_writer.get_extra_info('peername')
        if client_process is not None:
            return f'PID `{client_process.pid}`'
        elif client_peername is not None:
            return f'Socket `{client_peername}`'
        return 'Unknown'

    async def _handle_connection(self, client_reader: asyncio.StreamReader, client_writer: asyncio.StreamWriter):
        """
        Await a websocket to dedicate it for that connection and a _ConnectionHandler to manage communications between them
        """
        current_task = asyncio_current_task()
        with object_in_list(current_task, self._active_handle_connection_tasks):
            connection_identity_display_name = self._get_connection_identity_display_name(client_writer)
            self._logger.debug('Serving new connection from %s', connection_identity_display_name)
            websocket = None
            try:
                while websocket is None:
                    try:
                        websocket = await asyncio.wait_for(self._get_websocket(), timeout=5)
                        # We notify the remote that this websocket is about to be used
                        await websocket.send_str(WEBSOCKET_CONNECTION_START_MESSAGE)
                    except asyncio.TimeoutError:
                        # We stop waiting for websocket every few seconds to make sure the tunnel still works
                        # and the client is still await.
                        if not self.running or client_writer.transport.is_closing():
                            return
                    except ConnectionError:
                        # This might occurred when a queued websocket has been pending long enough, which means it got closed by
                        # the remote / proxy and we'll know about it only on the first message.
                        # The queue will be filled up with a fresh websocket so we try again.
                        websocket = None
                        continue
                connection = _ConnectionHandler(websocket=websocket, connection_reader=client_reader, connection_writer=client_writer, logger=self._logger)
                with object_in_list(connection, self._connections):
                    self._logger.debug('Start serving %s', connection_identity_display_name)
                    await connection.run_until_eof()
            except asyncio.CancelledError:
                self._logger.debug('Abort handling connection to %s', connection_identity_display_name)
            except Exception:
                self._logger.exception('Failed to serve new connection. Closing connection')
                raise
            finally:
                self._logger.debug('Close connection from %s', connection_identity_display_name)
                client_writer.close()
                if websocket:
                    await websocket.close()


class OutputTunnel(Tunnel):
    def __init__(self, exit_address, exit_port, entrance_address=None, entrance_port=None, logger: logging.Logger = None,
                 stop_tunnel_on_remote_callback: Callable[[bool], Coroutine] = None, stopped_callback: CallableOfCoroutine = None):
        """
        This is basically the tunnel's exit while InputTunnel is the entrance. See Tunnel class doc or this modules
        doc above for further understanding.
        """
        super().__init__(entrance_address, entrance_port, exit_address, exit_port, logger, stop_tunnel_on_remote_callback, stopped_callback)
        self._running_task: asyncio.Task = None
        self._serving_connections_tasks = []

    async def health_check(self):
        return await super().health_check() and not self._running_task.done()

    def get_local_socket(self):
        return self.get_exit_socket()

    def get_remote_socket(self):
        return self.get_entrance_socket()

    async def start(self):
        """
        Start processing new websockets and create connections
        """
        await super().start()
        self._running_task = asyncio.ensure_future(self._run())

    async def stop(self, stop_remote_tunnel=True, force=False):
        """
        Stop receiving new websockets and close all existing connections
        """
        try:
            if self._running_task is not None:
                with contextlib.suppress(asyncio.CancelledError):
                    self._running_task.cancel()
                    await self._running_task
        except Exception:
            self._logger.exception('An error occurred while stopping tunnel exit handler:')
        finally:
            await super().stop(stop_remote_tunnel=stop_remote_tunnel, force=False)
        if self._serving_connections_tasks:
            # super().stop() close the connections, and therefore those should be closed right after
            await asyncio.gather(*self._serving_connections_tasks)

    async def _serve_new_connection(self, websocket: WebsocketType):
        """
        Serve a new connection with the given websocket until eof reached.
        Then, mark the websocket_done_event
        """
        # We don't want to establish connections with the exit address and port if the websocket we received
        # is still queued on the remote InputTunnel. Therefore, we wait for an initial message before connecting.
        message: aiohttp.WSMessage = await websocket.receive()
        if websocket.closed:
            # Websocket was closed, we don't even start a connection
            return
        if message.type != aiohttp.WSMsgType.TEXT and message.data == WEBSOCKET_CONNECTION_START_MESSAGE:
            self._logger.error('Invalid first message: %s', message.data)
        try:
            reader, writer = await asyncio.open_connection(host=self._exit_address,
                                                           port=self._exit_port)
            connection = _ConnectionHandler(websocket=websocket, connection_reader=reader, connection_writer=writer,
                                            logger=self._logger)
            with object_in_list(connection, self._connections):
                await connection.run_until_eof()
        except ConnectionRefusedError:
            # No one is listening on the exit address and port, we close the tunnel normally.
            self._logger.debug('Connection Refused for `%s:%s`. Closing websocket', self._exit_address, self._exit_port)
        finally:
            await websocket.close()

    async def _run(self):
        """
        Wait for new websocket to arrive and open a connection object to (self._exit_address, self._exit_port).
        Then wait until either the connection or the websocket reach eof before marking the websocket as unneeded
        """
        while self.running:
            websocket = await self._get_websocket()
            task = asyncio.ensure_future(self._serve_new_connection(websocket))
            task_in_list_until_done(task, self._serving_connections_tasks)



from typing import Dict, Any

from aiohttp import web
from colorama import Fore, Style

from .. import __version__
from ..common import utils, const, security, exceptions, auth
from ..common.exceptions import NETunnelDestinationNotAllowed, NETunnelInvalidProxy
from .schemas import StaticTunnelSchema, PeerSchema
from .peer import Peer
from .config import NETunnelConfiguration
from .client_handler import ChannelHandler

import os
import json
import http
import logging
import asyncio
import argparse
import importlib
import functools
import contextlib


SECRET_STRING = "******"


def exceptions_to_json_response(async_route_method):
    """
    Convert exceptions to json error responses.
    http status codes can be bind to an exception class. Every unknown exception is considered internal_server_error (500)
    See netunnel.common.exceptions.NETunnelServerError
    :param async_route_method: A route method which accepts aiohttp aiohttp.web.Request object
    """
    @functools.wraps(async_route_method)
    async def wrap(self: 'NETunnelServer', request: web.Request, *args, **kwargs):
        try:
            return await async_route_method(self, request, *args, **kwargs)
        except exceptions.NETunnelServerError as err:
            return self._error_json(str(err), status=err.status_code.value)
        except Exception as err:
            self._logger.exception('The following exception raised for route `%s`:', request.rel_url)
            return self._error_json(str(err), status=http.HTTPStatus.INTERNAL_SERVER_ERROR.value)
    return wrap


def authenticated(async_route_method):
    """
    Validate that a request is authorized. Return Forbidden if not
    :param async_route_method: A route method which accepts aiohttp aiohttp.web.Request object
    """
    @functools.wraps(async_route_method)
    async def wrap(self: 'NETunnelServer', request: web.Request, *args, **kwargs):
        if await self._auth_server.is_authenticated(request):
            return await async_route_method(self, request, *args, **kwargs)
        self._logger.debug('Unauthorized request made to the route `%s`', request.rel_url)
        return web.HTTPForbidden()
    return wrap


class NETunnelServer:
    def __init__(self, config_path=None, host='127.0.0.1', port=4040, logger=None,
                 auth_server: auth.NETunnelServerAuth = None, secret_key=None):
        """
        NETunnel server. Listen on host:port and serve clients with tunnels.
        :param config_path: Path to the configuration file. If None, the server won't keep state
        :param host: Address to listen on
        :param port: Port to listen on
        :param logger: Optional logger
        :param auth_server: Instance of subclass of netunnel.common.auth.NETunnelServerAuth that will be used for authentication
        :param secret_key: A secret_key to use instead of generating and storing one in the configuration file
        """
        self._config_path = config_path
        self._host = host
        self._port = port
        self._logger = logger or utils.get_logger(__name__)
        self.channel_handlers: Dict[const.ChannelId, ChannelHandler] = {}
        self._channel_id_counter = 0
        # This lock is used to prevent duplications of channel_ids when multiple requests received at once
        self._channel_handlers_lock = asyncio.Lock()
        self._app = web.Application()
        self._runner: web.AppRunner = None
        self._auth_server: auth.NETunnelServerAuth = auth_server or auth.ServerNoAuth()
        self._setup_routes()
        self._config: NETunnelConfiguration = None
        # mapping of peer_id to Peer object
        self._peers: Dict[int, Peer] = {}
        # This lock is used when registering new peer to prevent id duplications
        self._registering_peer_lock = asyncio.Lock()
        # Used to prevent static tunnel's local port duplications on creation
        self._creating_static_tunnel_lock = asyncio.Lock()
        # This is used to encrypt and decrypt sensitive information
        self._secret_key = secret_key
        self._encryptor: security.Encryptor = None
        self._peer_schema = PeerSchema()
        self._static_tunnel_schema = StaticTunnelSchema()

    def _setup_routes(self):
        self._app.add_routes([
            web.get('/version', self.get_version),
            web.post('/channels', self.create_channel),  # Creates a channel
            web.post('/authenticate', self._auth_server.authenticate),
            web.get('/channels/{channel_id:\d+}/connect', self.serve_channel),  # Connect the channel (websocket)
            web.post('/channels/{channel_id:\d+}/tunnels', self.post_tunnel),  # Creates a tunnel
            web.delete('/channels/{channel_id:\d+}/tunnels/{tunnel_id:\d+}', self.delete_tunnel),  # Delete a tunnel
            web.get('/channels/{channel_id:\d+}/tunnels/{tunnel_id:\d+}/connect', self.websocket_to_tunnel),  # Feed a tunnel with a websocket
            web.get('/peers', self.list_peers),
            web.post('/peers', self.register_peer),
            web.get('/peers/{peer_id:\d+}', self.get_peer),
            web.post('/peers/{peer_id:\d+}', self.update_peer),
            web.delete('/peers/{peer_id:\d+}', self.delete_peer),
            web.get('/peers/{peer_id:\d+}/static_tunnels', self.list_peer_static_tunnels),
            web.post('/peers/{peer_id:\d+}/static_tunnels', self.create_peer_static_tunnels),
            web.get('/peers/{peer_id:\d+}/static_tunnels/{static_tunnel_id:\d+}', self.get_peer_static_tunnel),
            web.delete('/peers/{peer_id:\d+}/static_tunnels/{static_tunnel_id:\d+}', self.delete_peer_static_tunnel),
            web.get('/config/http-proxy', self.get_default_http_proxy),  # Return the default http proxy settings
            web.post('/config/http-proxy', self.set_default_http_proxy),  # Set the default http proxy
            web.post('/config/factory-reset', self.factory_reset)  # DANGEROUS - Erase and recreate the configurations
        ])

    def _get_decrypted_proxy_settings(self):
        """
        Return a tuple of the global proxy settings in the format (url, username, password)
        after decrypting the username / password if necessary
        """
        http_proxy_config: dict = self._config.get('http_proxy') or {}
        proxy_url = http_proxy_config.get('proxy_url')
        proxy_username = http_proxy_config.get('username')
        proxy_password = http_proxy_config.get('password')
        proxy_username = proxy_username and self._encryptor.decrypt_string(proxy_username)
        proxy_password = proxy_password and self._encryptor.decrypt_string(proxy_password)
        return proxy_url, proxy_username, proxy_password

    async def _setup_config(self):
        """
        Setup NETunnel configuration.
        If a configuration path was given, generate configuration from it
        """
        self._config = await NETunnelConfiguration.create(self._config_path)

    async def _setup_peers(self):
        """
        Setup peers and static tunnels that are stored in the configuration file.
        We do not need to verify them, as we trust that even if they aren't available right now, they will be
        """
        for peer_config in self._config['peers']:
            self._logger.info('Setup peer `%s`', peer_config['name'])
            peer = await self._load_peer_from_config(peer_config, verify_connectivity=False)
            self._peers[peer.id] = peer

    async def _detach_peers(self):
        """
        Disconnect static tunnels and stop peers.
        """
        # We cannot have new peers while we're detaching to avoid race conditions
        async with self._registering_peer_lock:
            for peer in list(self._peers.values()):
                self._logger.info('Disconnect peer `%s`', peer.name)
                await peer.delete_static_tunnels()
                del self._peers[peer.id]

    async def _shutdown_client_handlers(self):
        """
        Disconnect all connected clients.
        """
        # We convert to list because the items on self.channel_handlers are being removed after each loop
        for channel_id, channel_handler in list(self.channel_handlers.items()):
            self._logger.debug('Closing client handler `%s`', channel_id)
            await channel_handler.close()

    async def _setup_encryptor(self):
        """
        Initialize the secret key which will be used to encrypt and decrypt sensitive data.
        ** If the key was provided during initializing, we keep it only in memory **
        """
        if self._secret_key:
            self._encryptor = security.Encryptor(self._secret_key)
        elif 'NETUNNEL_SECRET_KEY' in os.environ:
            self._encryptor = security.Encryptor(os.environ['NETUNNEL_SECRET_KEY'])
        elif 'secret_key' in self._config:
            self._encryptor = security.Encryptor(self._config['secret_key'])
        else:
            self._logger.info('Generating default secret-key')
            secret_key = security.Encryptor.generate_key().decode()
            self._encryptor = security.Encryptor(secret_key)
            self._config['secret_key'] = secret_key
            await self._config.save()

    async def _start_web_server(self):
        self._logger.info('Starting server on %s:%s', self._host, self._port)
        self._runner = web.AppRunner(self._app)
        await self._runner.setup()
        site = web.TCPSite(self._runner, self._host, self._port)
        await site.start()

    async def _stop_web_server(self):
        if self._runner is not None:
            await self._runner.cleanup()

    async def _load_peer_from_config(self, peer_config, verify_connectivity=True) -> Peer:
        """
        Load a peer from a dictionary
        :param peer_config: settings of the peer to load in the format of dict
        :param verify_connectivity: Whether to verify connectivity to the peer
        """
        proxy_url, proxy_username, proxy_password = self._get_decrypted_proxy_settings()
        peer_load = self._peer_schema.load3(peer_config)
        # Once marshmallow supports asyncio in @post_load decorators, consider moving it there
        try:
            peer_load['auth'] = await self._auth_server.get_client_for_peer(**peer_load.pop('auth_data'))
        except (TypeError, ValueError) as err:
            raise exceptions.NETunnelServerBadRequest(str(err))
        # We strip the static_tunnels as they are initialized separately
        static_tunnels = peer_load.pop('static_tunnels', [])
        ssl = False if self._config['allow_unverified_ssl_peers'] else None
        peer = Peer(**peer_load, proxy_url=proxy_url, proxy_username=proxy_username, proxy_password=proxy_password,
                    ssl=ssl, logger=self._logger)
        if verify_connectivity:
            await peer.verify_connectivity()
        for static_tunnel in static_tunnels:
            await peer.add_static_tunnel(**static_tunnel, verify_connectivity=verify_connectivity)
        return peer

    @staticmethod
    def _error_json(err_message: str, data: Any = None, status=500, **json_response_kwargs):
        response = {
            'error': err_message
        }
        if data is not None:
            response['data'] = data
        return web.json_response(response, status=status, **json_response_kwargs)

    def _get_peer_by_id(self, peer_id) -> Peer:
        """
        Return peer object by peer_id. If peer not exists, raises NETunnelServerNotFound
        """
        if peer_id not in self._peers:
            raise exceptions.NETunnelServerNotFound(f'Peer `{peer_id}` was not found')
        return self._peers[peer_id]

    def _get_peer_by_name(self, peer_name) -> Peer:
        """
        Return peer object by name. If peer not exists, raises NETunnelServerNotFound
        :param peer_name: name of the peer to return
        """
        for peer in self._peers.values():
            if peer_name == peer.name:
                return peer
        raise exceptions.NETunnelServerNotFound(f'Peer `{peer_name}` was not found')

    def _get_peer_config_by_id(self, peer_id):
        """
        Return the config object of the peer by id. If peer not exists, raises NETunnelServerNotFound
        """
        for peer in self._config['peers']:
            if peer['id'] == peer_id:
                return peer
        raise exceptions.NETunnelServerNotFound(f'Peer `{peer_id}` was not found in the config')

    def _get_used_static_tunnels_local_ports(self):
        """
        Return a list of all the used local ports by static tunnels from all the peers.
        """
        used_ports = []
        for peer in self._peers.values():
            for tunnel in peer.static_tunnels:
                used_ports.append(tunnel.tunnel_local_port)
        return used_ports

    async def start(self):
        """
        Start netunnel server on host:port and start all static tunnels.
        """
        await self._setup_config()
        await self._setup_encryptor()
        await self._start_web_server()
        await self._setup_peers()

    async def stop(self):
        await self._detach_peers()
        await self._shutdown_client_handlers()
        self._logger.info('Stopping web application')
        await self._stop_web_server()

    @authenticated
    async def get_version(self, request: web.Request):
        """
        Return the running NETunnel version
        """
        return web.Response(text=__version__)

    @authenticated
    @exceptions_to_json_response
    async def create_channel(self, request: web.Request):
        """
        Creates a ChannelHandler object and return the channel_id to the client
        """
        data = await request.json()
        client_version = data['version']
        async with self._channel_handlers_lock:
            self._channel_id_counter += 1
            channel_id = self._channel_id_counter
        self.channel_handlers[channel_id] = ChannelHandler(channel_id=channel_id, client_version=client_version, logger=self._logger)
        # We send the version to support backward compatibility if needed
        return web.json_response({'channel_id': channel_id, 'version': __version__})

    @authenticated
    async def serve_channel(self, request: web.Request):
        """
        Creates a websocket to a channel and start serving it.
        Once we finish serving the channel, we delete it.
        """
        channel_id = int(request.match_info['channel_id'])
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        await self.channel_handlers[channel_id].serve(ws)
        del self.channel_handlers[channel_id]
        return ws

    def _verify_tunnel_destination_is_allowed(self, exit_address: str, exit_port: int):
        """
        This function purpose is to make sure a tunnel is not going to access a non-approved destination in the server network.
        It should be in used whenever the server opens a tunnel that its exit is on the server.
        The allowed destinations are set in the configuration under 'allowed_tunnel_destinations', and expected to be
        in the format of {"ip": "port1,port2,port3"}. "*" may be used to allow all ports.
        """
        allowed_tunnel_destinations = self._config.get('allowed_tunnel_destinations')
        if allowed_tunnel_destinations is None:
            raise RuntimeError("NETunnel configurations is missing 'allowed_tunnel_destinations'")
        if exit_address not in allowed_tunnel_destinations:
            raise NETunnelDestinationNotAllowed(f'Address {exit_address} is not allowed as a tunnel destination')
        allowed_ports_for_address = [port.strip() for port in allowed_tunnel_destinations[exit_address].split(',')]
        if str(exit_port) not in allowed_ports_for_address and "*" not in allowed_ports_for_address:
            raise NETunnelDestinationNotAllowed(f'Port {exit_address}:{exit_port} is not allowed as a tunnel destination')

    @authenticated
    @exceptions_to_json_response
    async def post_tunnel(self, request: web.Request):
        """
        Creates a tunnel.
        Each tunnel belong to a ChannelHandler and identified by a tunnel_id.
        return that tunnel_id to the client.
        """
        channel_id = int(request.match_info['channel_id'])
        data = await request.json()
        try:
            if data.pop('reverse'):
                self._logger.info('Opening Server-To-Client tunnel. Listen on: %s:%s', data['entrance_address'], data['entrance_port'])
                tunnel_id = await self.channel_handlers[channel_id].create_server_to_client_tunnel(**data)
            else:
                self._verify_tunnel_destination_is_allowed(data['exit_address'], data['exit_port'])
                self._logger.info('Opening Client-To-Server tunnel. Connecting to: %s:%s', data["exit_address"], data["exit_port"])
                tunnel_id = await self.channel_handlers[channel_id].create_client_to_server_tunnel(**data)
        except (RuntimeError, NETunnelDestinationNotAllowed) as err:
            return self._error_json(str(err))
        return web.json_response({'tunnel_id': tunnel_id})

    @authenticated
    @exceptions_to_json_response
    async def delete_tunnel(self, request: web.Request):
        """
        Delete a tunnel.
        Unlike channels which always have a single alive websocket, tunnels can have many websockets or no websockets
        at a given time, this request is the way for the client to tell the server to delete a tunnel
        """
        channel_id = int(request.match_info['channel_id'])
        tunnel_id = int(request.match_info['tunnel_id'])
        data = await request.json()
        force = data.get('force', False)
        self._logger.debug('Deleting tunnel `%s` on channel `%s`', tunnel_id, channel_id)
        try:
            await self.channel_handlers[channel_id].delete_tunnel(tunnel_id, force)
        except KeyError:
            return self._error_json(f'tunnel_id `{tunnel_id}` does not exists')
        return web.json_response()

    @authenticated
    async def websocket_to_tunnel(self, request: web.Request):
        """
        Generates a new websocket to an existing tunnel
        """
        channel_id = int(request.match_info['channel_id'])
        tunnel_id = int(request.match_info['tunnel_id'])
        self._logger.debug('Serving new websocket for tunnel `%s` on channel `%s`', tunnel_id, channel_id)
        ws = utils.EventWebSocketResponse()
        await ws.prepare(request)
        await self.channel_handlers[channel_id].serve_new_connection(ws, tunnel_id)
        return ws

    @authenticated
    @exceptions_to_json_response
    async def factory_reset(self, request: web.Request):
        """
        THIS CANNOT BE UNDONE, PLEASE USE WITH CAUTION

        Recreate the configuration file from the defaults.
        The web won't be restarted, so all currently connected channels won't be affected unless
        `disconnect_clients` is True.
        """
        data = await request.json()
        await self._detach_peers()
        await self._config.recreate()
        await self._setup_encryptor()
        if data.get('disconnect_clients', False):
            await self._shutdown_client_handlers()
        return web.json_response()

    @authenticated
    @exceptions_to_json_response
    async def get_default_http_proxy(self, request: web.Request):
        response = self._config.get('http_proxy') or {}
        # Credentials are encrypted
        if response.get('username'):
            response['username'] = SECRET_STRING
            response['password'] = SECRET_STRING
        return web.json_response(response)

    @authenticated
    @exceptions_to_json_response
    async def set_default_http_proxy(self, request: web.Request):
        """
        Set the default http proxy that will be used when static tunnels are established.
        If there are existing static tunnel, they will be reset to apply the new proxy
        """
        data = await request.json()
        if 'proxy_url' not in data:
            return self._error_json("Missing 'proxy_url' for proxy")
        proxy_url = data['proxy_url']
        # A test url that will be used to verify the proxy
        test_url = data.get('test_url') or self._config.get('http_proxy_test_url')
        username = data.get('username')
        password = data.get('password')
        check_proxy = data.get('check_proxy', True)

        if proxy_url is None:
            self._config['http_proxy'] = None
            self._logger.info('Setting empty default proxy url')
        else:
            if check_proxy:
                try:
                    await utils.verify_proxy(proxy_url=proxy_url, test_url=test_url, username=username, password=password)
                except (ValueError, NETunnelInvalidProxy) as err:
                    self._logger.exception('Failed to set default http proxy:')
                    return self._error_json(str(err))
            self._logger.info('Settings default proxy url to `%s`', proxy_url)
            self._config['http_proxy'] = {
                'proxy_url': proxy_url,
                'username': username and self._encryptor.encrypt_string(username),
                'password': password and self._encryptor.encrypt_string(password)
            }
        await self._config.save()
        for peer in self._peers.values():
            await peer.set_new_proxy(proxy_url=proxy_url, proxy_username=username, proxy_password=password)
        return await self.get_default_http_proxy(request)

    @authenticated
    @exceptions_to_json_response
    async def list_peers(self, request: web.Request):
        """
        Return a list of registered peers
        """
        name_filter = request.rel_url.query.get('name')
        peers_to_return = self._config['peers']
        if name_filter:
            peers_to_return = filter(lambda peer: peer['name'] == name_filter, peers_to_return)
        return web.json_response(list(peers_to_return))

    @authenticated
    @exceptions_to_json_response
    async def get_peer(self, request: web.Request):
        """
        Return a peer by id
        """
        peer_id = int(request.match_info['peer_id'])
        peer = self._get_peer_by_id(peer_id)
        return web.json_response(self._peer_schema.dump3(peer))

    @authenticated
    @exceptions_to_json_response
    async def register_peer(self, request: web.Request):
        """
        Register a new peer
        """
        data = await request.json()
        # Verify data
        errors = self._peer_schema.validate(data)
        if errors:
            return self._error_json(f'Failed to register peer: {errors}', status=http.HTTPStatus.BAD_REQUEST.value)

        # Verify unique peer name
        new_peer_name = data['name']
        with contextlib.suppress(exceptions.NETunnelServerNotFound):
            if self._get_peer_by_name(new_peer_name):
                return self._error_json(f"Peer `{new_peer_name}` already exists", status=http.HTTPStatus.BAD_REQUEST.value)

        async with self._registering_peer_lock:
            # Generate an id for the peer
            data['id'] = max(self._peers.keys()) + 1 if self._peers else 1
            new_peer = await self._load_peer_from_config(data)

            # Save new peer
            self._peers[new_peer.id] = new_peer
            new_peer_config = self._peer_schema.dump3(new_peer)
            self._config['peers'].append(new_peer_config)
        await self._config.save()
        return web.json_response(new_peer_config)

    @authenticated
    @exceptions_to_json_response
    async def update_peer(self, request: web.Request):
        """
        Update peer data
        """
        peer_id = int(request.match_info['peer_id'])
        peer = self._get_peer_by_id(peer_id)
        peer_config: dict = self._get_peer_config_by_id(peer_id)
        data = await request.json()

        # We load only the updatable fields
        changeable_fields = ('name', 'target_netunnel_url', 'auth_data')
        partial_schema = PeerSchema(only=changeable_fields)
        fields_to_update = partial_schema.load3(data, partial=changeable_fields)
        if 'auth_data' in fields_to_update:
            # Once marshmallow supports asyncio in @post_load decorators, consider moving it there
            try:
                fields_to_update['auth'] = await self._auth_server.get_client_for_peer(**fields_to_update.pop('auth_data'))
            except (TypeError, ValueError) as err:
                raise exceptions.NETunnelServerBadRequest(str(err))

        # Validate unique name.
        with contextlib.suppress(exceptions.NETunnelServerNotFound):
            # This will raise and ignored if no peer with this name exists
            existing_peer = self._get_peer_by_name(fields_to_update.get('name'))
            # We allow to 'update' the name to the same name since its the expected behavior for POST requests
            if existing_peer is not peer:
                return self._error_json(f"Peer `{fields_to_update['name']}` already exists", status=http.HTTPStatus.BAD_REQUEST.value)
        if fields_to_update.get('name') in [p.name for p in self._peers.values() if p != peer]:
            return self._error_json(f"Peer `{fields_to_update['name']}` already exists")
        # Validate connectivity by creating a temporary peer
        if 'target_netunnel_url' in fields_to_update or 'auth' in fields_to_update:
            proxy_url, proxy_username, proxy_password = self._get_decrypted_proxy_settings()
            # We use the new authentication data or/and new url so we'll know if they're valid before updating the real peer
            url = fields_to_update.get('target_netunnel_url') or peer.target_netunnel_url
            auth = fields_to_update.get('auth') or peer.auth
            ssl = False if self._config['allow_unverified_ssl_peers'] else None
            temp_peer = Peer(id=-1, name='temp', target_netunnel_url=url, auth=auth,
                             proxy_url=proxy_url, proxy_username=proxy_username, proxy_password=proxy_password,
                             ssl=ssl, logger=self._logger)
            await temp_peer.verify_connectivity()

        # Modify the peer object
        if 'target_netunnel_url' in fields_to_update or 'auth' in fields_to_update:
            # This will trigger the static tunnels to be recreated
            await peer.update_settings(new_url=fields_to_update.get('target_netunnel_url'), new_auth=fields_to_update.get('auth'))
        if 'name' in fields_to_update:
            new_name = fields_to_update['name']
            self._logger.info('Renaming peer `%s` to `%s`', peer.name, new_name)
            peer.name = new_name
        peer_config.update(self._peer_schema.dump3(peer))
        await self._config.save()
        return web.json_response(peer_config)

    @authenticated
    @exceptions_to_json_response
    async def delete_peer(self, request: web.Request):
        """
        Deletes a peer by an id
        """
        peer_id = int(request.match_info['peer_id'])
        peer = self._get_peer_by_id(peer_id)
        peer_config = self._get_peer_config_by_id(peer_id)
        await peer.delete_static_tunnels()
        self._config['peers'].remove(peer_config)
        await self._config.save()
        del self._peers[peer.id]
        return web.json_response()

    @authenticated
    @exceptions_to_json_response
    async def list_peer_static_tunnels(self, request: web.Request):
        """
        Return a list of static tunnels of a peer
        """
        peer_id = int(request.match_info['peer_id'])
        peer = self._get_peer_by_id(peer_id)
        return web.json_response(self._static_tunnel_schema.dump3(peer.static_tunnels, many=True))

    @authenticated
    @exceptions_to_json_response
    async def create_peer_static_tunnels(self, request: web.Request):
        """
        Creates a static tunnel to peer
        """
        peer_id = int(request.match_info['peer_id'])
        peer = self._get_peer_by_id(peer_id)
        data: dict = await request.json()
        async with self._creating_static_tunnel_lock:
            # Validate data
            if 'tunnel_local_port' in data:
                return self._error_json("Tunnel's local_port is generated by the netunnel server")
            elif 'id' in data:
                return self._error_json("Tunnel's id is generated by the netunnel server")
            data['tunnel_local_port'] = utils.get_unused_port(min_port=const.MIN_STATIC_TUNNEL_LOCAL_PORT,
                                                              max_port=const.MAX_STATIC_TUNNEL_LOCAL_PORT,
                                                              exclude_ports=self._get_used_static_tunnels_local_ports())
            errors = self._static_tunnel_schema.validate(data)
            if errors:
                return self._error_json(f'Failed to create static tunnel for peer `{peer.name}`: {errors}', status=http.HTTPStatus.BAD_REQUEST.value)
            # We're using the schema's load to generate the default values and emit unnecessary parameters
            static_tunnel = await peer.add_static_tunnel(**self._static_tunnel_schema.load3(data))
            # We're dumping from the object so we'll have auto-generated fields (ID for example) as well
            static_tunnel_settings = self._static_tunnel_schema.dump3(static_tunnel)
            peer_config = self._get_peer_config_by_id(peer_id)
            peer_config['static_tunnels'].append(static_tunnel_settings)
            await self._config.save()
        return web.json_response(static_tunnel_settings)

    @authenticated
    @exceptions_to_json_response
    async def get_peer_static_tunnel(self, request: web.Request):
        """
        Return a static tunnel of a peer by an id
        """
        peer_id = int(request.match_info['peer_id'])
        static_tunnel_id = int(request.match_info['static_tunnel_id'])
        peer = self._get_peer_by_id(peer_id)
        static_tunnel = peer.get_static_tunnel(static_tunnel_id)
        return web.json_response(self._static_tunnel_schema.dump3(static_tunnel))

    @authenticated
    @exceptions_to_json_response
    async def delete_peer_static_tunnel(self, request: web.Request):
        """
        Delete a static tunnel of a peer by an id
        :param request:
        :return:
        """
        peer_id = int(request.match_info['peer_id'])
        static_tunnel_id = int(request.match_info['static_tunnel_id'])
        peer = self._get_peer_by_id(peer_id)
        static_tunnel = peer.get_static_tunnel(static_tunnel_id)
        peer_config = self._get_peer_config_by_id(peer_id)
        static_tunnel_config = self._static_tunnel_schema.dump3(static_tunnel)
        await peer.delete_static_tunnel(static_tunnel_id)
        peer_config['static_tunnels'].remove(static_tunnel_config)
        await self._config.save()
        return web.json_response()


def main():
    parser = argparse.ArgumentParser(description='Run a netunnel server')
    parser.add_argument('-c', '--config-path', help='Configuration file path to use or generate if missing. Defaults to stateless mode')
    parser.add_argument('-d', '--debug', help="increase log verbosity to debug mode", dest='loglevel', action="store_const",
                        const=logging.DEBUG, default=logging.INFO)
    parser.add_argument('-p', '--port', type=int, help="Port number to listen on", default=4040)
    parser.add_argument('-s', '--secret-key',
                        help="A passphrase used for encryption which will not be stored in the configuration file")
    parser.add_argument('--hostname', help="IP address to listen on", default='127.0.0.1')
    parser.add_argument('--auth-plugin', help="Plugin to use for authentication. (e.g. <module_path>.<auth_class_name>)")
    parser.add_argument('--auth-data', default='{}', help="A json dump string of the data required by the authentication plugin")
    args = parser.parse_args()
    logger = utils.get_logger(name='netunnel_server', level=args.loglevel)
    auth_server = None
    if args.auth_plugin:
        module, class_name = args.auth_plugin.rsplit('.', maxsplit=1)
        auth_server_class = getattr(importlib.import_module(module), class_name)
        auth_server = auth_server_class(**json.loads(args.auth_data))
    if args.config_path is None:
        print(f'{Fore.YELLOW}The server is running in stateless mode. Use --config-path to generate a config file{Style.RESET_ALL}')
    server = NETunnelServer(config_path=args.config_path, logger=logger, port=args.port, auth_server=auth_server,
                            secret_key=args.secret_key)
    loop = asyncio.get_event_loop()
    try:
        asyncio.ensure_future(server.start())
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        loop.close()


if __name__ == "__main__":
    main()
